"""
BaseInterface Module
====================

This module provides a base interface for defining default classes and options. It includes functionality for setting default parameters, updating interface options, and retrieving updated options.

"""


from .BaseInterface import BaseInterface
from .NonlinearHeat import NonlinearHeat

__all__ = ['BaseInterface', 'NonlinearHeat']


import numpy as np
from ..interface.BaseInterface import BaseInterface as bs
from .Iteral import Iteral as Int
import scipy.sparse as sp

_DEFAULT_OPTS = {
    'stop': 'normal', 
    'maxit': 1e4, 
    'tol': 1e-8, 
    'verbose': False,
    'mode': None, 
    'method': 'normal',
    }


def _returns(x, fnorm, dxnorm, mode, method):
    """
    Helper function to format the return value based on the mode.

    Parameters
    ----------
    x : array_like
        Root obtained after iterations.
    fnorm : float
        Norm of the residuals.
    dxnorm : float
        Norm of the correction.
    mode : str
        Mode of the output ('full', 'partial', None).
    method : str
        Stopping method used.

    Returns
    -------
    tuple
        Tuple containing the formatted output based on the mode.
    """
    
    if mode == 'full':
        return x, dxnorm, fnorm
    elif mode == 'partial':
        if method == 'normal': 
            return x, dxnorm 
        else: 
            return x, fnorm
        
    elif ((mode == None) or (mode == False)):
        return x, dxnorm, fnorm

@bs.set_defaults(default_cls = Int, default_opts=_DEFAULT_OPTS)
def newton(x, f, df, **kwargs):
    """
    Perform Newton iterations to find the root of a given function.

    Parameters
    ----------
    x : array_like
        Initial guess for the root.
    f : callable
        Function to evaluate the residuals.
    df : callable
        Function to evaluate the Jacobian matrix.
    **kwargs : dict
        Additional keyword arguments for customization.
        tol : float, optional
            Tolerance for convergence. Default is 1e-8.
        maxit : int, optional
            Maximum number of iterations. Default is 10000.
        verbose : bool, optional
            If True, prints iteration information. Default is False.
        mode : str, optional
            Mode of the output ('full', 'partial', None).
    
    Returns
    -------
    root, correction_norm, residuals_norm : tuple
        If mode is 'full'
    root, correction_norm, (residuals_norm) : tuple
        if mode is 'partial'. The residuals_norm are returned if method is not 'normal' 
    
    root, iterations, correction_norm, residuals_norm: tuple
        if mode is None

    Raises
    ------
    ValueError
        If the maximum number of iterations or tolerance is not a positive integer.

    Examples
    --------
    Basic usage:
    >>> root, iterations, norm_correction = newton(2.0, lambda x: x**2 - 4, lambda x: 2 * x, verbose=True)
    >>> print("Root:", root, "Iterations:", iterations, "Norm of correction:", norm_correction)
        
    Usage with kwargs provided as a dictionary:
    >>> kwargs = {'verbose': True, 'tol': 1e-6, 'maxit': 20}
    >>> root, norm_correction = newton(3.0, lambda x: x**3 - 27, lambda x: 3 * x**2, **kwargs)
    >>> print("Root:", root, "Norm of correction:", norm_correction)
    """
    _opts = bs.opts(**kwargs)
    verbose = _opts['verbose']
    mode = _opts['mode']
    tol = _opts['tol']
    maxit = int(_opts['maxit'])
    stop = _opts['stop']
    
    if maxit <= 0 or not isinstance(maxit, int):
        raise ValueError("Maximum number of iterations 'maxit' must be a positive integer.")
    if tol <= 0:
        raise ValueError("Tolerance 'tol' must be a positive value.")
    
    if verbose:
        print('------ Newton Method summary ------')
        print(f'tollerence: {tol}')
        print(f'maximum iter: {maxit}')
        print(f'stopping criteria: {stop}')
        print(f'starting guess: {x}')
    
        print('------ Start iteration ------')
        
    normdx = []
    normf = []
    for iter in range(maxit):
        f_value = f(x)
        df_value = df(x)
        
        if sp.issparse(df_value) and sp.issparse(df_value):
            dx = -sp.linalg.spsolve(df_value, f_value)
            dxnorm = sp.linalg.norm(dx)
        else:
            dx = -np.linalg.solve(df_value, f_value)
            dxnorm = np.linalg.norm(dx)
        
        x += dx.reshape((len(dx),))
        
        if sp.issparse(df_value) and sp.issparse(df_value):
            fnorm = sp.linalg.norm(f(x))
        else:
            fnorm = np.linalg.norm(f(x))
            
        normdx.append(dxnorm)
        normf.append(fnorm)
        
        if  (verbose != 0 and verbose != False) and (iter % verbose == 0):
            print(f"Newton status at iteration {iter + 1}: ||dx|| = {dxnorm} and ||F|| = {fnorm}")
        
        if (fnorm < tol and stop == 'residual-check'):
            if verbose:
                print(f'Newton converged in {iter + 1} iterations with ||F|| = {fnorm}')
            return _returns(x, fnorm=fnorm, dxnorm=normdx, mode=mode, method=stop)
        
        if (dxnorm < tol and stop == 'normal'):
            if verbose:
                print(f'Newton converged in {iter + 1} iterations with ||dx|| = {dxnorm}')
            return _returns(x, fnorm=fnorm, dxnorm=normdx, mode=mode, method=stop)
        

    if verbose:
        print(f'Newton did not converge within {maxit} iterations')

    return _returns(x, fnorm=fnorm, dxnorm=normdx, mode=mode, method=stop)



import numpy as np
from scipy.linalg import lu, inv, solve
from numpy.linalg import qr
from ..interface.BaseInterface import BaseInterface as bs
from .Iteral import Iteral as Int
import scipy.sparse as sp

_DEFAULT_OPTS = {
        'stop': 'matrix',
        'maxit': 1e4,
        'tol': 1e-8,
        'method': 'standard',
        'verbose': True,
    }


    
def _Z(V, A = None, P=None, L=None, U=None, method = 'standard'):
    if method == 'LU':
        y = solve(inv(U), P.dot(V))
        z = solve(inv(L), y)
    else:
        z = A @ V
    return z

@bs.set_defaults(default_cls = Int, default_opts=_DEFAULT_OPTS)
def osim(A, V, **kwargs):
    """
    Orthogonal Subspace Iteration Method (OSIM).

    Args:
    A : numpy.ndarray
        The matrix to compute the eigenvalues and eigenvectors for.
    V : numpy.ndarray
        Initial guess of eigenvectors.

    Keyword Arguments:
        tol : float, optional
            Tolerance for convergence (default: 1e-8).
        maxit : int, optional
            Maximum number of iterations (default: 100).
        stop : str, optional
            Stopping criteria for convergence. Options are 'eig' (default), 'matrix', or 'residual'.
        method : str, optional
            Method for solving linear systems. Options are 'LU' (default) or any method supported by scipy.linalg.lu.
        verbose : bool, optional
            If True, prints iteration information (default: True).

    Returns:
    numpy.ndarray
        Matrix of eigenvectors.
    numpy.ndarray
        Matrix of transformed eigenvectors.
    tuple
        Tuple containing eigenvalues at each iteration.

    Notes:
    This function implements the Orthogonal Subspace Iteration Method (OSIM) to compute eigenvectors and eigenvalues of a matrix A.

    Examples:
    >>> import numpy as np
    >>> from byma.interal import osim
    >>> A = np.array([[1, 0], [0, 1]])
    >>> V = np.array([[1], [0]])
    >>> V, BV, iter = osim(A, V, tol=1e-8, maxit=1000, stop='eig', method='LU')

    You can also pass keyword arguments using a dictionary. For example:
    >>> kwargs = {'tol': 1e-8, 'maxit': 1000, 'stop': 'eig', 'method': 'LU'}
    >>> V, BV, iter = osim(A, V, **kwargs)

    You can also pass keyword arguments using two separate dictionaries for parameters and interface. For example:
    >>> parameters = {'tol': 1e-8, 'maxit': 1000, 'stop': 'eig', 'method': 'LU'}
    >>> interface = {'verbose': True}
    >>> V, BV, iter = sosim(A, V, parameters=parameters, interface=interface)
    """
    
    # Check if the number of rows of V matches the number of columns of A
    if V.shape[0] != A.shape[1]:
        raise ValueError("Number of rows of V must match the number of columns of A.")
    
    _opts = bs.opts(**kwargs)
    verbose = _opts['verbose']
    tol = _opts['tol']
    maxit = int(_opts['maxit'])
    stop = _opts['stop']
    method = _opts['method']
    
    if verbose:
        print('------ OSIM initialization summary ------')
        print(f'tollerence: {tol}')
        print(f'maximum iter: {maxit}')
        print(f'stopping criteria: {stop}')
        print(f'Linear system solving method: {method}')
    
        print('------ Start iteration ------')
    
    if method == 'LU':
        if sp.issparse(A):
            P, L, U = lu(A.toarray())
        else:
            P, L, U = lu(A)
    else:
        P = L = U = None

    B = lambda V: V.T @ _Z(V = V, A = A, P = P, L = L, U = U, method = method)

    iter = []
    for n in range(maxit):
        Zn = _Z(V = V, A = A, P = P, L = L, U = U, method = method)
        BV = B(V = V)
        eig = np.diag(BV)
        if verbose != False: 
            if (n % verbose == 0):
                print(f"Eigenvalues at n = {n}: {eig}")
        else:
            if (n % 5000 == 0):
                print(f"iteration n = {n}")
        
        if n > 0:
            iter.append(eig)
        
        V, _ = qr(Zn)

        if stop == 'eig':
            if n > 1:
                if abs(np.linalg.norm(eig - iter[n - 1])) < tol:
                    print(f'The Orthogonal Subspace Method has converged in {n} iterations.')
                    break
            else:
                pass
        elif stop == 'matrix':
            
            if n > 1:
                if np.linalg.norm(BV - B(V)) < tol:
                    print(f'The Orthogonal Subspace Method has converged in {n} iterations.')
                    break
        
        elif stop == 'residual':
            
            if np.linalg.norm(A @ V - V @ B(V)) < tol:
                print(f'The Orthogonal Subspace Method has converged in {n} iterations.')
                break
    
        if n >= maxit:
            print('The Orthogonal Subspace Method has not converged')
        
    return V, BV, tuple(iter)
        
        

    
    


import numpy as np
from ..interface.BaseInterface import BaseInterface as bs
from .Numy import Numy as Nu
from ..iteral import newton
import scipy.sparse as sp

    
class _root():
        
    def g(u, z, t, f, dt=0.01, i=1):
        return u - i*dt * f(x = u, t = t) - z
        
    def dg(u, z, t, df, dt=0.01, i=1):
        n = len(z)
        return sp.identity(n) - i*dt * df(x = u, t = t)
    
_DEFAULT_OPTS = {
    'stop': 'normal', 
    'maxit': 1e4, 
    'tol': 1e-8, 
    'verbose': False,
    'mode': None, 
    'method': 'backward',
    }

@bs.set_defaults(default_cls = Nu, default_opts=_DEFAULT_OPTS)
def euler(start=0, T=1, dt=0.01, x0=None, f=None, df=None, **kwargs):
    """
    Euler integration method.

    Parameters
    ----------
    start : float, optional
        Start time of integration (default: 0).
    T : float, optional
        End time of integration (default: 1).
    dt : float, optional
        Time step for integration (default: 0.01).
    x0 : numpy.ndarray, optional
        Initial condition (default: None).
    f : callable, optional
        Function representing the derivative of the system (default: None).
    df : callable, optional
        Function representing the derivative of the derivative of the system (default: None).
    **kwargs : dict, optional
        Additional keyword arguments.

    Returns
    -------
    sol : list
        List containing the solution at each time step.
    iter_list : list
        List containing the norm of the change in solution at each iteration.
    time : numpy.ndarray
        Array containing the time steps.

    Raises
    ------
    ValueError
        If x0, f, or df is None.
        If an invalid continuation method is provided.

    Notes
    -----
    This function implements the Euler integration method for solving ordinary differential equations.

    Examples
    --------
    >>> import numpy as np
    >>> from byma.numy import euler
    >>> def f(x, t):
    ...     return -x * t
    >>> sol, iter_list, time = euler(f=f, x0=1, T=1, dt=0.1)
    """
    _opts = bs.opts(**kwargs)
    verbose = _opts['verbose']
    method = _opts['method']
    
    # netwon default parameters for euler method
    kwargs['mode'] = 'partial'
    kwargs['stop'] = 'normal'

    # Set some parameters
    bs.check_none(x0, f, df)
    time = np.arange(start, T, dt)
    
    if verbose:
        print('------ Euler Method summary ------')
        print(f'Starting time: {start}')
        print(f'Ending time: {T}')
        print(f'Time step: {dt}')
        print(f'method: {method}')
    
        print('------ Start iteration ------')
        
        
    # Time loop 
    sol = []
    iter_list = []
    func = _root
    for i,t in enumerate(time):

        f1 = lambda x: func.g(u=x, z=x, t=t, f=f, dt=dt, i=i)
        df1 = lambda x: func.dg(u=x, z=x, t=t, df=df, dt=dt, i=i)
        
        if method == 'backward':
            x0, dxnorm = newton(x = x0, f = f1, df = df1, **kwargs)
        elif method == 'foward':
            raise ValueError("Foward Euler method is not yet avialible")
        elif method == 'trapezoidal':
            raise ValueError("Trapezoidal method is not yet avialible")
        else:
            raise ValueError("Invalid continuation method. Choose either 'backward', 'foward' or 'trapezoidal'.")

        if (verbose != 0 and verbose != False) and (i % verbose == 0):
            print(f'Solution at time {t} is {x0}')
            
        sol.append(x0.copy())
        
        iter_list.append(dxnorm)
        
    return sol, iter_list, time


"""
``byma.numy``
================

The ByMa basic numerical methods sub-package.

Functions present in byma.numy are listed below.


Continuation
--------------------------

   euler
   

Exceptions
----------

   NuMyError

"""
# To get sub-modules
from . import _numy
from ._numy import *

__all__ = _numy.__all__.copy()


import matplotlib.pyplot as plt
import numpy as np

    

def plot_scale(x, y, label, scale, style=None):
    """
    Plot data with specified scale and style.

    Parameters
    ----------
    x : array-like
        Data points for the x-axis.
    y : array-like
        Data points for the y-axis.
    label : str
        Label for the data.
    scale : str
        Scale for the plot. Options: 'normal', 'loglog', 'xlog', 'ylog'.
    style : dict or None, optional
        Dictionary specifying line style. If None, default style is used.

    Returns
    -------
    None
    """
    if scale == 'loglog':
        plt.loglog(x, y, label=label)
    elif scale == 'xlog':
        plt.semilogx(x, y, label=label)
    elif scale == 'ylog':
        plt.semilogy(x, y, label=label)
    else:
        plt.plot(x, y, label=label)

def plot(x, y, **kwargs):
    """
    Function that create plots of different kinds. With this function it is possible to plots n numbers of 
    function in the same figure quickly and with high personalization.

    Parameters
    ----------
    x : array-like
        Array of x values.
    y : array-like
        Array of y values.
    **kwargs : dict, optional
        settings (dict): Overall plot settings.
            title (str): Title of the plot.
            xlabel (str): Label for the x-axis.
            ylabel (str): Label for the y-axis.
            label (str):  Label for the 1st function.
            label{i} (str): Label for the ith function from i=2.
            x{i} (array-like): Array of x values for the ith function from i=2.
            y{i} (array-like): Array of y values for the ith function from i=2.
            save_title (str): File name to save the plot.
            save_path (str): Path to save the plot. If None, the plot will be saved in the current directory.
            scale (str): Scale for the plot. Options: 'normal', 'loglog', 'xlog', 'ylog'.
        style (dict or list of dicts): Line style(s) for the plot. (Not yet working)

    Returns
    -------
    None
    """
    settings = kwargs.pop('settings', {})
    settings.update(kwargs)  # Incorporate any additional kwargs into settings

    scale = settings.get('scale', 'normal')
    style = settings.get('style', None)

    plt.figure()
    plt.title(settings.get('title', 'Plot'))
    plt.xlabel(settings.get('xlabel', 'x'))
    plt.ylabel(settings.get('ylabel', 'y'))

    plot_scale(x, y, label=settings.get('label', 'y'), scale=scale, style=style)

    for i in range(2, 10):
        y_i = settings.get(f'y{i}', None)
        if y_i is not None:
            label_i = settings.get(f'label{i}', f'y{i}')
            style_i = style[i-2] if isinstance(style, list) else style
            plot_scale(x, y_i, label=label_i, scale=scale, style=style_i)
        else:
            break

    plt.legend()
    plt.show()

    save_title = settings.get('save_title', settings.get('title', 'Plot'))
    save_path = settings.get('save_path', None)

    # Save the plot if requested
    if save_title:
        if save_path is None:
            save_path = '.'
        plt.savefig(f"{save_path}/{save_title}.png")
        print(f"Plot saved at: {save_path}/{save_title}.png")
        
        
def plot_numerical_error(n, func, solve_func, save_title=None, save_path=None, **kwargs):
    """
    Plot the numerical error between the exact and numerical solutions and print the maximum error.

    Parameters
    ----------
    n : int
        Number of grid points.
    func : object
        Instance of the class containing the exact solution.
    solve_func : function
        Function to solve the system.
    save_title : str, optional
        File name plot.
    save_path : str, optional
        Path to save the plot. If None, the plot will be saved in the current directory.
    **kwargs : dict, optional
        Additional keyword arguments to customize the plot and title for saving.

    Returns
    -------
    None
    """

    # Solve the sparse linear system
    numerical_solution = solve_func(n=n, func=func)

    # Generate the x values for plotting
    x_values = np.linspace(0, 1, n + 1)[1:]

    # Calculate the exact solution at each x value
    exact_solution = func.exact(x_values)

    # Calculate the error between the numerical and exact solutions
    error = abs(numerical_solution - exact_solution)

    # Plot the error
    plt.plot(x_values, error, label=kwargs.get('label', 'Error'))
    plt.xlabel(kwargs.get('x_label', 'x'))
    plt.ylabel(kwargs.get('y_label', 'Error'))
    plt.title(kwargs.get('title', 'Error in Numerical Solution'))
    plt.legend()
    
    
    # Save the plot if requested
    if save_title:
        if save_path is None:
            save_path = '.'
        plt.savefig(f"{save_path}/{save_title}.png")
        print(f"Plot saved at: {save_path}/{save_title}.png")

    
    # Show the plot
    plt.show()

    # Print the maximum error
    max_error = np.max(np.abs(error))
    print(f"Maximum error: {max_error}")


import numpy as np
import scipy.sparse as sp


class NonlinearHeat():
    """
    Damped stationary heat equation with discretization using finite difference method.

    The equation is given by:

    .. math::
        u_{xx} + \mu(u - \\frac{1}{3}u^3)   x \in [0,1]
        u(0) = 1, u(1) = 0

    The linearized version is:

    .. math::
        u_{xx} + \mu u    x \in [0,1]
        u(0) = 1, u(1) = 0

    Functions:

    - `GL1(x, mu)`: Checks if x is a zero of the right-hand side.
    - `fGL1(u, n, mu)`: Computes the resulting sparse matrix.
    - `linGL1(n)`: Computes a 1d float vector for the linearized version.
    - `JacGL1(u, n, mu)`: Computes the Jacobian matrix.

    """
    
    def __init__(self):
        pass

    def GL1(self, x, mu):
        """
        Computes the finite difference method for the damped stationary heat equation.

        Parameters
        ----------
        x : numpy.ndarray
            1D vector.
        mu : float
            Scalar parameter.

        Returns
        -------
        numpy.ndarray
            Result of the computation.

        """
        A = self.linGL1(n = len(x) + 1)
        return A @ x + mu * (x - x**3)

    def fGL1(self, **kwargs):
        """
        Computes the resulting sparse matrix for the damped stationary heat equation.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments including 'u', 'n', and 'mu'.

        Returns
        -------
        numpy.ndarray
            Resulting sparse matrix.

        """
        u = kwargs.get('u')
        n = kwargs.get('n')
        assert type(n) == int, 'integer only'
        assert np.ndim(u) == 1, '1d only'
        
        mu = kwargs.get("mu", np.pi**2)
        A = self.linGL1(n = n)
        vec = mu*u  - (mu /3) * (u ** 3)
        
        return (A @ u + vec)
    
    def linGL1(self, n, **kwargs):
        """
        Computes a 1D float vector for the linearized version of the heat equation.

        Parameters
        ----------
        n : int
            Number of grid points.

        Returns
        -------
        numpy.ndarray
            1D float vector.

        """
        assert type(n) == int, 'integer only'
        
        top = np.full(n - 2, 1)
        mid = np.full(n - 1, -2)
        bot = np.full(n - 2, 1)
        h = 1/(n - 1)
        
        return (h**(-2)) * sp.csr_matrix(sp.diags([ top , mid , bot ] , [1 , 0 , -1]))
    
    def JacGL1(self, **kwargs):
        """
        Computes the Jacobian matrix for the damped stationary heat equation.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments including 'u', 'n', and 'mu'.

        Returns
        -------
        numpy.ndarray
            Jacobian matrix.

        """
        u = kwargs.get('u')
        n = kwargs.get('n', int(5))
        mu = kwargs.get("mu", np.pi)
        assert type(n) == int, 'integer only'
        # assert np.ndim(u) == 1, '1d only'
        
        A = self.linGL1(n=n)
        
        # top = np.full(n - 2, 1)
        # bot = np.full(n - 2, - 1)
        # B = ((2*h)**(-1)) * sp.csr_matrix(sp.diags([ top , bot ] , [1 , -1]))
        # vec = mu * B @ (1 - u**2)
       
        # ttop = np.full(n - 3, 1)
        # top = np.full(n - 2, - 2)
        # bot = np.full(n - 2, - 2)
        # bbot = np.full(n - 3, - 1)
        # C = ((2*(h**3))**(-1)) * sp.csr_matrix(sp.diags([ ttop, top , bot, bbot ] , [2, 1 , -1, -2]))
        # print(C == B @ A)
        
        # return (B @ A) @ u + vec
        return A + mu * (np.eye(n - 1) - np.diag(u**2))


import functools

class BaseInterface:
    '''Defines a base interface'''

    def __init__(self, default_cls, default_opts, **kwargs):
        self.cls = default_cls
        self.default_opts = default_opts
        self.kwargs = kwargs

    @staticmethod
    def set_defaults(default_cls, default_opts={}):
        """
        Decorator for setting default interface and parameters.

        Parameters
        ----------
        default_cls : obj
            The default class instance.
        default_opts : dict, optional
            Default options for the interface (default is an empty dictionary).

        Returns
        -------
        callable
            A decorator function that sets default interface and parameters.
        """
        def decorator(func):
            @functools.wraps(func)
            def wrapper(**kwargs):
                
                _opts = default_cls()._opts.copy()
                _opts.update(default_opts)
                
                kwargs['default_opts'] = _opts
                kwargs['default_cls'] = default_cls

                return func(**kwargs)
            return wrapper
        return decorator

    @staticmethod
    def opts(**kwargs):
        """
        Method for setting options for the interface.

        Parameters
        ----------
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        dict
            A dictionary containing updated interface and parameters.
        """
        # Construct instance of default class with the updated interface and parameters 
        usr_opts = kwargs.pop('usr_opts', None) 
        _opts = kwargs.pop('default_opts') 
        

        if usr_opts is not None:
            _opts.update(usr_opts)
            
        _opts.update(kwargs)

        return dict(_opts)
    
    def check_none(*args):
        """
        Check if any of the arguments are None.

        Parameters
        ----------
        *args : tuple
            Arbitrary number of arguments to check.

        Returns
        -------
        bool
            True if any argument is None, False otherwise.

        Raises
        ------
        ValueError
            If any argument is None, raises ValueError with the names of the None arguments.

        
        """
        none_indices = [i for i, arg in enumerate(args) if arg is None]
        if none_indices:
            none_names = [args[i].__name__ if hasattr(args[i], '__name__') else str(args[i]) for i in none_indices]
            raise ValueError(f"The following arguments are None: {', '.join(none_names)}")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import pathlib
import sys
import os
import re
import importlib
from docutils import nodes
from recommonmark.parser import CommonMarkParser


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ByMa'
copyright = '2024, Lorenzo Zambelli'
author = 'Lorenzo Zambelli'


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

sys.path.insert(0, os.path.abspath('../sphinxext'))

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.apidoc',
    'autoapi.extension',
    'sphinx.ext.todo',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.mathjax',
    'sphinx_design',
]

source_parsers = {
    '.md': CommonMarkParser,
}

source_suffix = ['.rst', '.md']

# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
#

import byma
import byma.nuby

# The short X.Y version (including .devXXXX, rcX, b1 suffixes if present)
# version = re.sub(r'(\d+\.\d+)\.\d+(.*)', r'\1\2', byma.__version__)
version = re.sub(r'(\d+\.\d+\.\d+).*', r'\1', byma.__version__)
# version = re.sub(r'(\.dev\d+).*?$', r'\1', version)
# The full version, including alpha/beta/rc tags.
release = byma.__version__
print("%s %s" % (version, release))

# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "autolink"

autoapi_dirs = [
    "../../byma",
    "../../byma/nuby",
    "../../byma/interface",
    "../../byma/iteral",
    "../../byma/pyplot"
]
autoapi_type = "python"
autoapi_template_dir = "_templates/autoapi"
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "imported-members",
]
autoapi_keep_files = True
autodoc_typehints = "signature"
templates_path = ['_templates']
exclude_patterns = []

# -----------------------------------------------------------------------------
# Autosummary
# -----------------------------------------------------------------------------

autosummary_generate = True



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# Set up the version switcher.  The versions.json is stored in the doc repo.
if os.environ.get('CIRCLE_JOB', False) and \
        os.environ.get('CIRCLE_BRANCH', '') != 'main':
    # For PR, name is set to its ref
    switcher_version = os.environ['CIRCLE_BRANCH']
elif ".dev" in version:
    switcher_version = "devdocs"
else:
    switcher_version = f"{version}"

html_theme_options = {
#   "logo": {
#       "image_light": "_static/",
#       "image_dark": "_static/",
#   },
  "github_url": "https://gitlab.com/ByteMath/python/ByMa",
  "collapse_navigation": True,
  # Add light/dark mode and documentation version switcher:
}
html_title = "%s v%s Manual" % (project, version)
html_last_updated_fmt = '%b %d, %Y'
# html_css_files = [".css"]
html_context = {"default_mode": "light"}
html_use_modindex = True
html_copy_source = False
html_domain_indices = False
html_file_suffix = '.html'






"""
``byma.nuby``
================

The ByMa numerical bifurcation functions rely on 

Functions present in byma.nuby are listed below.


Continuation
--------------------------

   step
   cont
   

Exceptions
----------

   NuByError

"""
# To get sub-modules
from . import _nuby
from ._nuby import *

__all__ = _nuby.__all__.copy()


from setuptools import setup, find_packages

DESCRIPTION = 'Numerical Mathematics Package'
LONG_DESCRIPTION = 'A Numerical Mathematics Packages with an easier and clean user interface.'

# Setting up
setup(
    name="byma",
    include_package_data=True,
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    author="@b64-Lorenzo (Lorenzo Zambelli)",
    author_email="<bytemath@lorenzozambelli.it>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION + '\n\n' + open('CHANGELOG.txt').read(),
    packages=find_packages(),
    install_requires=[
        'numpy>=1.1', 
        'scipy>=1.1', 
        'matplotlib>=3.6'
        ],
    keywords=['python', 'scientific', 'numerical', 'bifurcation'],
    classifiers=[ "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import numpy as np
import scipy.sparse as sp
from ..iteral import newton
from ..interface.BaseInterface import BaseInterface as bs
from .Bifurcation import Bifurcation as Bf

_DEFAULT_OPTS = {
    'maxit_cont' : 1000,
    'verbose': True,
    'mode': None, 
    'method': 'normal',
    }


def _returns(x, branch, target, mode, dxnorms, mu0):
    """
    Helper function to format the return value based on the mode.

    Parameters
    ----------
    x : array_like
        Root obtained after iterations.
    fnorm : float
        Norm of the residuals.
    dxnorm : float
        Norm of the correction.
    mode : str
        Mode of the output ('full', 'partial', None).
    method : str
        Stopping method used.

    Returns
    -------
    tuple
        Tuple containing the formatted output based on the mode.
    """
    
    if target == None:
        if mode == 'full':
            return branch, dxnorms, mu0 
        elif mode == 'partial':
            return branch, mu0
        else:
            return x, mu0
    else:
        if mode == 'full':
            return branch, dxnorms
        elif mode == 'partial':
            return branch
        else:
            return x


def step(x, f, df, dfmu, dx, dmu, **kwargs):
        """
    Perform one step of the continuation.

    :param x: array_like
        Current state.
    :param f: callable
        Function to evaluate the Jacobian matrix with respect to state variable x.
    :param df: callable
        Function to evaluate the Jacobian matrix with respect to state variable x.
    :param dfmu: callable
        Function to evaluate the Jacobian matrix derivative with respect to parameter mu.
    :param dx: array_like
        Current tangent with respect to state variable.
    :param dmu: float or array_like
        Incremental change in the parameter value.
    :param kwargs: dict
        Additional keyword arguments.

    :return: tuple
        A tuple containing the updated state x and the norm of the correction and the derivative of x w.r.t parameter

    :description:
    This function performs one step of the continuation, updating the state variable x and the tangent dx using the provided functions for evaluating the Jacobian matrix and its derivative with respect to the parameter.
    """
        
        # Predictor
        x += dmu * dx
        kwargs['mode'] = 'partial'
        kwargs['stop'] = 'normal'
  
        # Corrector 
        x, dxnorm = newton(x = x, f = f, df = df, **kwargs)

        # Compute the tangent
        dfx1 = df(x = x)
        dfmu1 = dfmu(x = x)
        if sp.issparse(dfx1) and sp.issparse(dfmu1):
            dx = sp.linalg.spsolve(dfx1, -dfmu1)
            
        else:
            dx = np.linalg.solve(dfx1, -dfmu1)
        
        return x, dxnorm, dx

@bs.set_defaults(default_cls = Bf, default_opts=_DEFAULT_OPTS)
def cont(x0, dx0, start, f, df, dfmu, dmu=None, target=None, **kwargs):
    """
    Perform a continuation in parameter value from a starting value to a target value or until the maximum iteration is met, with constant step size.

    This function performs a continuation in parameter space from a starting value to a target value, or until the maximum iteration is met, adjusting the state variable x along the way. The continuation is carried out using the provided functions for evaluating the Jacobian matrix and its derivative with respect to the parameter.

    :param x0: array_like
        Initial state.
    :param dx0: array_like
        Initial tangent with respect to state variable.
    :param start: float
        Starting parameter value.
    :param f: callable
        Function to evaluate the Jacobian matrix with respect to state variable x.
    :param df: callable
        Function to evaluate the Jacobian matrix with respect to state variable x and parameter mu.
    :param dfmu: callable
        Function to evaluate the Jacobian matrix derivative with respect to parameter mu.
    :param dmu: float or array_like, optional
        Incremental change in the parameter value for each iteration. If None and target is None, raises ValueError.
    :param target: float or None, optional
        Target parameter value. If None, continuation is performed until maxit_con iterations.
    :param kwargs: dict
        Additional keyword arguments for customization.
            maxit_con : int, optional
                Maximum number of continuation steps. Default is 1000.
            method : str, optional
                Continuation method ('normal' or 'pseudo-arclength'). Default is 'normal'.
            mode : str, optional
                Return mode ('partial' or 'full'). Default is 'partial'.
            Other keyword arguments : Additional parameters specific to the step function used internally.

    :return: tuple or array_like
        Depending on the mode specified in kwargs, returns either a tuple or an array.
            - In 'partial' mode, returns a tuple containing the final state x and the final parameter value mu.
            - In 'full' mode, returns an array containing all the states x, an array of the norm of the correction at each step, and the final parameter value mu (if target is None).

    :raises:
        ValueError: If either 'dmu' or 'target' should be not 'None' but are not provided.
        ValueError: If the provided continuation method is invalid. Choose either 'normal' or 'pseudo-arclength'.

    Examples
    =============
        >>> # Define the functions df and dfmu
        >>> def df(x, mu0):
        >>>     # Compute the Jacobian matrix with respect to state variable x and parameter mu
        >>>     pass
        >>> def dfmu(x, mu0):
        >>>     # Compute the Jacobian matrix derivative with respect to parameter mu
        >>>     pass
        >>> 
        >>> # Define the initial state and tangent
        >>> x0 = np.array([1.0, 2.0])
        >>> dx0 = np.array([0.1, 0.1])
        >>> 
        >>> # Perform continuation from start value to target value
        >>> start = 0.0
        >>> target = 1.0
        >>> result = cont(x0, dx0, start, df, dfmu, target=target, maxit_con=1000, method='normal', mode='full')
        >>> print(result)
    """
    _opts = bs.opts(**kwargs)
    mode_con = _opts['mode']
    method = _opts['method']
    maxit_con = _opts['maxit_cont']
    verbose = _opts['verbose']
    
    dx = dx0.copy()
    mu = start

    # Set some parameters
    if (target is None) and (dmu is None):
        raise ValueError("Either 'dmu' or 'target' should be not 'None'")
    
    else:
        mus = np.linspace(mu, target, maxit_con)
        dmu = mus[1] - mu
        
        
    # Perform the continuation
    branch = []
    dxnorms = []
    
    if verbose:
        print('------ Continuation Method summary ------')
        print(f'starting solution: {x0}')
        print(f'starting parameter: {start}')
        print(f'maximum iter: {maxit_con}')
        print(f'method: {method}')
        if isinstance(dmu, float):
            print(f'Constand dmu: {dmu}')
    
        print('------ Start iteration ------')

    for j in range(maxit_con):
        
        if target is None:
            mu0 = mu0 + dmu if isinstance(dmu, float) else mu0 + dmu[j]
        else:
            mu0 = mus[j].copy()

        f1 = lambda x: f(x = x, mu = mu0)
        df1 = lambda x: df(x = x, mu = mu0)
        dfmu1 = lambda x: dfmu(x = x, mu = mu0)
        
        if method == 'normal':
            x0, dxnorm, dx = step(x = x0, f=f1, df=df1, dfmu=dfmu1, dx=dx, dmu=dmu, **kwargs)
        elif method == 'pseudo-arclength':
            raise ValueError("Pseudo-arclength'conitnuation is not yet avialible")
        else:
            raise ValueError("Invalid continuation method. Choose either 'normal' or 'pseudo-arclength'.")

        branch.append(x0.copy())
        dxnorms.append(dxnorm)

    mode = mode_con
    x = x0.copy()
    return _returns(x, branch, target, mode, dxnorms, mu0)


"""
ByMa: A scientific computing package for Python
================================================

Documentation is available in the docstrings and
online at 

Subpackages
-----------
Using any of these subpackages requires an explicit import. For example,
``import byma.nuby``.

::

 nuby                         --- Numerical Bifurcation Analysis Tools
 iteral                       --- Tools for iterative algorithms
 pyplot                       --- Tools for plotting functions
 interface                    --- Interface functions

"""

import importlib as _importlib

try:
    from importlib.metadata import version  # for Python 3.8+
except ImportError:
    from pkg_resources import get_distribution, DistributionNotFound
    try:
        __version__ = get_distribution(__name__).version
    except DistributionNotFound:
        # Package is not installed
        __version__ = 'unknown'
else:
    __version__ = version(__name__)



submodules = [
    'nuby',
    'iteral',
    'pyplot',
]

__all__ = submodules + [
    'test',
]


def __dir__():
    return __all__


def __getattr__(name):
    if name in submodules:
        return _importlib.import_module(f'byma.{name}')
    else:
        try:
            return globals()[name]
        except KeyError:
            raise AttributeError(
                f"Module 'byma' has no attribute '{name}'"
            )